@@ -1,7 +1,7 @@
 /**************************************************************************
  *   cut.c  --  This file is part of GNU nano.                            *
  *                                                                        *
- *   Copyright (C) 1999-2011, 2013-2019 Free Software Foundation, Inc.    *
+ *   Copyright (C) 1999-2011, 2013-2020 Free Software Foundation, Inc.    *
  *   Copyright (C) 2014 Mark Majeres                                      *
  *   Copyright (C) 2016, 2018, 2019 Benno Schulenberg                     *
  *                                                                        *
@@ -41,7 +41,7 @@ void do_deletion(undo_type action)
 		/* If the type of action changed or the cursor moved to a different
 		 * line, create a new undo item, otherwise update the existing item. */
 		if (action != openfile->last_action ||
-					openfile->current->lineno != openfile->current_undo->lineno)
+					openfile->current->lineno != openfile->current_undo->head_lineno)
 			add_undo(action, NULL);
 		else
 			update_undo(action);
@@ -120,7 +120,7 @@ void do_delete(void)
 		zap_text();
 	else
 #endif
-	do_deletion(DEL);
+		do_deletion(DEL);
 }
 
 /* Backspace over one character.  That is, move the cursor left one
@@ -132,12 +132,38 @@ void do_backspace(void)
 		zap_text();
 	else
 #endif
-	if (openfile->current != openfile->filetop || openfile->current_x > 0) {
+	if (openfile->current_x > 0 || openfile->current != openfile->filetop) {
 		do_left();
 		do_deletion(BACK);
 	}
 }
 
+/* Return FALSE when a cut command would not actually cut anything: when
+ * on an empty line at EOF, or when the mark covers zero characters, or
+ * (when test_cliff is TRUE) when the magic line would be cut. */
+bool is_cuttable(bool test_cliff)
+{
+	size_t from = (test_cliff) ? openfile->current_x : 0;
+
+	if ((openfile->current->next == NULL && openfile->current->data[from] == '\0'
+#ifndef NANO_TINY
+					&& openfile->mark == NULL) ||
+					(openfile->mark == openfile->current &&
+					openfile->mark_x == openfile->current_x) ||
+					(from > 0 && !ISSET(NO_NEWLINES) &&
+					openfile->current->data[from] == '\0' &&
+					openfile->current->next == openfile->filebot
+#endif
+					)) {
+#ifndef NANO_TINY
+		statusbar(_("Nothing was cut"));
+		openfile->mark = NULL;
+#endif
+		return FALSE;
+	} else
+		return TRUE;
+}
+
 #ifndef NANO_TINY
 /* Delete text from the cursor until the first start of a word to
  * the left, or to the right when forward is TRUE. */
@@ -205,11 +231,167 @@ void chop_next_word(void)
 {
 	openfile->mark = NULL;
 
-	if (is_cuttable(openfile->current_x > 0))
+	if (is_cuttable(TRUE))
 		chop_word(FORWARD);
 }
 #endif /* !NANO_TINY */
 
+/* Move all text between (top, top_x) and (bot, bot_x) from the current buffer
+ * into the cutbuffer. */
+void extract_segment(linestruct *top, size_t top_x, linestruct *bot, size_t bot_x)
+{
+	bool edittop_inside = (openfile->edittop->lineno >= top->lineno &&
+							openfile->edittop->lineno <= bot->lineno);
+#ifndef NANO_TINY
+	bool mark_inside = (openfile->mark &&
+						openfile->mark->lineno >= top->lineno &&
+						openfile->mark->lineno <= bot->lineno &&
+						(openfile->mark != top || openfile->mark_x >= top_x) &&
+						(openfile->mark != bot || openfile->mark_x <= bot_x));
+	bool same_line = (openfile->mark == top);
+
+	if (top == bot && top_x == bot_x)
+		return;
+#endif
+
+	/* Reduce the buffer to cover just the text that needs to be extracted. */
+	partition_buffer(top, top_x, bot, bot_x);
+
+	/* Subtract the number of characters in that text from the file size. */
+	openfile->totsize -= get_totsize(top, bot);
+
+	/* If the cutbuffer is currently empty, just move all the text directly
+	 * into it; otherwise, append the text to what is already there. */
+	if (cutbuffer == NULL) {
+		cutbuffer = openfile->filetop;
+		cutbottom = openfile->filebot;
+	} else {
+		/* Tack the data of the first line of the text onto the data of
+		 * the last line in the given buffer. */
+		cutbottom->data = charealloc(cutbottom->data,
+								strlen(cutbottom->data) +
+								strlen(openfile->filetop->data) + 1);
+		strcat(cutbottom->data, openfile->filetop->data);
+
+		/* Attach the second line of the text (if any) to the last line
+		 * of the buffer, then remove the now superfluous first line. */
+		cutbottom->next = openfile->filetop->next;
+		delete_node(openfile->filetop);
+
+		/* If there is a second line, make the reverse attachment too and
+		 * update the buffer pointer to point at the end of the text. */
+		if (cutbottom->next != NULL) {
+			cutbottom->next->prev = cutbottom;
+			cutbottom = openfile->filebot;
+		}
+	}
+
+	/* Since the text has now been saved, remove it from the file buffer. */
+	openfile->filetop = make_new_node(NULL);
+	openfile->filetop->data = copy_of("");
+	openfile->filebot = openfile->filetop;
+
+	/* Set the cursor at the point where the text was removed. */
+	openfile->current = openfile->filetop;
+	openfile->current_x = top_x;
+#ifndef NANO_TINY
+	/* If the mark was inside the partition, put it where the cursor now is. */
+	if (mark_inside) {
+		openfile->mark = openfile->current;
+		openfile->mark_x = openfile->current_x;
+	} else if (same_line)
+		/* Update the pointer to this partially cut line. */
+		openfile->mark = openfile->current;
+#endif
+
+	/* Glue the texts before and after the extraction together. */
+	unpartition_buffer();
+
+	renumber_from(openfile->current);
+
+	/* If the top of the edit window was inside the old partition, put
+	 * it in range of current. */
+	if (edittop_inside) {
+		adjust_viewport(STATIONARY);
+		refresh_needed = TRUE;
+	}
+
+	/* If the text doesn't end with a newline, and it should, add one. */
+	if (!ISSET(NO_NEWLINES) && openfile->filebot->data[0] != '\0')
+		new_magicline();
+}
+
+/* Meld the buffer that starts at topline into the current file buffer
+ * at the current cursor position. */
+void ingraft_buffer(linestruct *topline)
+{
+	size_t current_x_save = openfile->current_x;
+	/* Remember whether the current line is at the top of the edit window. */
+	bool edittop_inside = (openfile->edittop == openfile->current);
+#ifndef NANO_TINY
+	/* Remember whether mark and cursor are on the same line, and their order. */
+	bool right_side_up = (openfile->mark && mark_is_before_cursor());
+	bool same_line = (openfile->mark == openfile->current);
+#endif
+
+	/* Partition the buffer so that it contains no text, then delete it.*/
+	partition_buffer(openfile->current, openfile->current_x,
+						openfile->current, openfile->current_x);
+	delete_node(openfile->filetop);
+
+	/* Replace the current buffer with the passed buffer. */
+	openfile->filetop = topline;
+	openfile->filebot = topline;
+	while (openfile->filebot->next != NULL)
+		openfile->filebot = openfile->filebot->next;
+
+	/* Put the cursor at the end of the pasted text. */
+	openfile->current = openfile->filebot;
+	openfile->current_x = strlen(openfile->filebot->data);
+
+	/* When the pasted stuff contains no newline, adjust the cursor's
+	 * x coordinate for the text that is before the pasted stuff. */
+	if (openfile->filetop == openfile->filebot)
+		openfile->current_x += current_x_save;
+
+#ifndef NANO_TINY
+	/* When needed, refresh the mark's pointer and compensate the mark's
+	 * x coordinate for the change in the current line. */
+	if (same_line) {
+		if (!right_side_up) {
+			openfile->mark = openfile->filebot;
+			openfile->mark_x += openfile->current_x - current_x_save;
+		} else
+			openfile->mark = openfile->filetop;
+	}
+#endif
+
+	/* Add the number of characters in the copied text to the file size. */
+	openfile->totsize += get_totsize(openfile->filetop, openfile->filebot);
+
+	/* If we pasted onto the first line of the edit window, the corresponding
+	 * record has been freed, so... point at the start of the copied text. */
+	if (edittop_inside)
+		openfile->edittop = openfile->filetop;
+
+	/* Weld the pasted text into the surrounding content of the buffer. */
+	unpartition_buffer();
+
+	renumber_from(topline);
+
+	/* If the text doesn't end with a newline, and it should, add one. */
+	if (!ISSET(NO_NEWLINES) && openfile->filebot->data[0] != '\0')
+		new_magicline();
+}
+
+/* Meld a copy of the given buffer into the current file buffer. */
+void copy_from_buffer(linestruct *somebuffer)
+{
+	linestruct *the_copy = copy_buffer(somebuffer);
+
+	ingraft_buffer(the_copy);
+}
+
 /* Move the whole current line from the current buffer to the cutbuffer. */
 void cut_line(void)
 {
@@ -217,9 +399,9 @@ void cut_line(void)
 	 * head of this line to the head of the next line into the cutbuffer;
 	 * otherwise, move all of the text of this line into the cutbuffer. */
 	if (openfile->current != openfile->filebot)
-		extract(openfile->current, 0, openfile->current->next, 0);
+		extract_segment(openfile->current, 0, openfile->current->next, 0);
 	else
-		extract(openfile->current, 0,
+		extract_segment(openfile->current, 0,
 				openfile->current, strlen(openfile->current->data));
 
 	openfile->placewewant = 0;
@@ -237,10 +419,10 @@ void cut_to_eol(void)
 	 * the cutbuffer.  Otherwise, when not at the end of the buffer,
 	 * move the line separation into the cutbuffer. */
 	if (openfile->current_x < data_len)
-		extract(openfile->current, openfile->current_x,
+		extract_segment(openfile->current, openfile->current_x,
 				openfile->current, data_len);
 	else if (openfile->current != openfile->filebot) {
-		extract(openfile->current, openfile->current_x,
+		extract_segment(openfile->current, openfile->current_x,
 				openfile->current->next, 0);
 		openfile->placewewant = xplustabs();
 	}
@@ -255,7 +437,7 @@ void cut_marked(bool *right_side_up)
 	get_region((const linestruct **)&top, &top_x,
 				(const linestruct **)&bot, &bot_x, right_side_up);
 
-	extract(top, top_x, bot, bot_x);
+	extract_segment(top, top_x, bot, bot_x);
 
 	openfile->placewewant = xplustabs();
 }
@@ -263,7 +445,7 @@ void cut_marked(bool *right_side_up)
 /* Move all text from the cursor position to end-of-file into the cutbuffer. */
 void cut_to_eof(void)
 {
-	extract(openfile->current, openfile->current_x,
+	extract_segment(openfile->current, openfile->current_x,
 				openfile->filebot, strlen(openfile->filebot->data));
 }
 #endif /* !NANO_TINY */
@@ -350,30 +532,6 @@ void do_snip(bool copying, bool marked, bool until_eof, bool append)
 	refresh_needed = TRUE;
 }
 
-/* Return FALSE when a cut command would not actually cut anything: when
- * on an empty line at EOF, or when the mark covers zero characters, or
- * (when test_cliff is TRUE) when the magic line would be cut. */
-bool is_cuttable(bool test_cliff)
-{
-	if ((openfile->current->next == NULL && openfile->current->data[0] == '\0'
-#ifndef NANO_TINY
-					&& openfile->mark == NULL) ||
-					(openfile->mark == openfile->current &&
-					openfile->mark_x == openfile->current_x) ||
-					(test_cliff && openfile->current->data[openfile->current_x] == '\0' &&
-					((ISSET(NO_NEWLINES) && openfile->current == openfile->filebot) ||
-					(!ISSET(NO_NEWLINES) && openfile->current == openfile->filebot->prev))
-#endif
-					)) {
-#ifndef NANO_TINY
-		openfile->mark = NULL;
-#endif
-		statusbar(_("Nothing was cut"));
-		return FALSE;
-	} else
-		return TRUE;
-}
-
 /* Move text from the current buffer into the cutbuffer. */
 void cut_text(void)
 {
@@ -429,9 +587,10 @@ void copy_text(void)
 /* Cut from the current cursor position to the end of the file. */
 void cut_till_eof(void)
 {
-	if ((openfile->current == openfile->filebot && openfile->current->data[0] == '\0') ||
-				(!ISSET(NO_NEWLINES) && openfile->current->next == openfile->filebot &&
-				openfile->current->data[openfile->current_x] == '\0')) {
+	if (openfile->current->data[openfile->current_x] == '\0' &&
+				(openfile->current->next == NULL ||
+				(!ISSET(NO_NEWLINES) && openfile->current_x > 0 &&
+				openfile->current->next == openfile->filebot))) {
 		statusbar(_("Nothing was cut"));
 		return;
 	}
